# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from cms.utils.mail import send_mail
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from cms.models import Title

from workflows.models import Action


EDITOR, AUTHOR = 'editor', 'author'

SUBJECTS = {
    Action.REQUEST: {
        EDITOR: _('{project}: Change review requested')
    },
    Action.APPROVE: {
        EDITOR: _('{project}: Change review requested'),
        AUTHOR: _('{project}: Changes approved'),
    },
    Action.REJECT: {
        AUTHOR: _('{project}: Changes rejected'),
    },
    Action.CANCEL: {
        AUTHOR: _('{project}: Changes cancelled'),
    },
}


def send_action_mails(action, editor=None):
    """
    Returns `True` if any email has been sent.

    :param action: Action triggering mails
    :type action: Action
    :type editor: django.contrib.auth.models.AbstractUser
    :rtype: bool
    """
    sent = False

    if action.action_type not in SUBJECTS:
        return sent
    subjects = SUBJECTS[action.action_type]
    context = _context(action)

    if AUTHOR in subjects:
        subject = subjects[AUTHOR].format(**context)
        txt_template = 'workflows/emails/author_{}.txt'.format(action.action_type)
        to = get_to(action, to_user=action.get_author())
        send_mail(subject, txt_template, to, context=context)
        sent = True

    if EDITOR in subjects:
        subject = subjects[EDITOR].format(**context)
        to = get_to(action, to_user=editor)
        if to:
            txt_template = 'workflows/emails/editor_{}.txt'.format(action.action_type)
            send_mail(subject, txt_template, to, context=context)
            sent = True

    return sent


def _context(action):
    author = action.get_author()
    editor = action.user
    context = {
        'url': get_absolute_url(action.title),
        'author_name': get_name(author, default=_('author')),
        'editor_name': get_name(editor, default=_('editor')),
        'message': action.message,
        'project': getattr(settings, 'PROJECT_NAME', 'djangocms-workflows')
    }
    return context


def get_to(action, to_user=None):
    """
    :type to_user: django.contrib.auth.models.AbstractUser
    :param to_user:
    :return:
    """
    if to_user:
        return [to_user.email]
    return [user.email for user in action.next_mandatory_stage_editors()]


def get_absolute_url(title):
    """
    :type title: Title
    :param title:
    :return:
    """
    scheme = ['http', 'https'][getattr(settings, 'USE_HTTPS', False)]
    domain = title.page.site.domain
    path = title.page.get_absolute_url(language=title.language).lstrip('/')
    return '{}://{}/{}'.format(scheme, domain, path)


def get_name(user, default=None):
    try:
        name = user.get_full_name()
        assert name.strip() != ''
        return name
    except (AttributeError, AssertionError):
        try:
            return user.get_username()
        except AttributeError:
            return default or str(user)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from copy import copy
from lxml import etree

from cms.models import Page, Title
from cms.plugin_rendering import ContentRenderer

from django.contrib import messages
from django.core.urlresolvers import reverse
from django.http import Http404
from django.shortcuts import render, redirect, get_object_or_404
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from django.views.generic import TemplateView
from django.views.generic.edit import FormView
from lxml.html.diff import htmldiff, parse_html

from sekizai.context import SekizaiContext

from .email import send_action_mails
from .forms import ActionForm
from .models import Action, Workflow


NO_WORKFLOW = _('There is no workflow for this page and language.')
ACTIVE_REQUEST = _('There already is an active request for this page and language.')
NO_ACTIVE_REQUEST = _('There is no active request for this page and language.')
USER_NOT_ALLOWED = _('You are not allowed to approve or reject this request.')

# this closes the admin sideframe overlay and redirects to 'url' (in context)
CLOSE_FRAME = 'workflows/admin/action_confirm.html'


class ActionView(FormView):

    template_name = 'workflows/admin/action_form.html'
    form_class = ActionForm
    action_type = None
    admin_title = None
    admin_save_label = None
    confirm_message = _('Done')

    @cached_property
    def language(self):
        """
        Returns current language.

        :rtype: str
        """
        return self.args[1]

    @cached_property
    def page(self):
        """
        Returns current page which must be draft.

        :rtype: Page
        """
        pk = self.args[0]
        try:
            return Page.objects.get(
                pk=pk,
                publisher_is_draft=True,  # redundant, but ensuring we only deal with drafts
            )
        except Page.DoesNotExist:
            raise Http404

    @cached_property
    def title(self):
        """
        Returns title instance of current page/language.

        :rtype: Title
        """
        try:
            return self.page.title_set.get(language=self.language)
        except Title.DoesNotExist:
            raise Http404

    @cached_property
    def workflow(self):
        """
        Returns the appropriate workflow for the current title.

        :rtype: workflows.models.Workflow
        """
        return Workflow.get_workflow(self.title)

    @cached_property
    def user(self):
        """
        Returns the user calling the view.

        :rtype: django.contrib.auth.models.User
        """
        return self.request.user

    @cached_property
    def action_request(self):
        """
        Returns the initial request action of the current title's current action chain.

        :rtype: Action
        """
        return Action.get_current_request(self.title)

    @cached_property
    def stage(self):
        """
        Returns the workflow stage that will be associated with the action created by this view.
        That can be None in the case of the initial request which is never associated with a stage.

        :rtype: workflows.models.WorkflowStage
        """
        if not self.action_request:
            return None
        return self.action_request.last_action().get_next_stage(self.user)

    def validate(self):
        """Validates that this view can legally be called with all the current parameters.
        """
        if self.workflow is None:
            raise InvalidAction(NO_WORKFLOW)

    def get_success_url(self):
        """
        Url to redirect to after successful post.

        :rtype: str
        """
        return self.page.get_draft_url(language=self.language)

    def get_failed_url(self):
        """
        Url to redirect to after failed validation.

        :rtype: str
        """
        return self.request.META.get('HTTP_REFERER', self.get_success_url())

    def get_form_url(self):
        """
        Url to submit form to.

        :rtype: str
        """
        return self.request.path

    def dispatch(self, request, *args, **kwargs):
        try:
            self.validate()
        except InvalidAction as e:
            messages.error(request, e.message)
            return redirect(self.get_failed_url())
        return super(ActionView, self).dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        """
        Kwargs to initialize form with. The form needs some extra information
        as the actual Action creation is handled by the form.

        :rtype: dict
        """
        kwargs = super(ActionView, self).get_form_kwargs()
        kwargs.update({
            'title': self.title,
            'stage': self.stage,
            'workflow': self.workflow,
            'action_type': self.action_type,
            'request': self.request,
        })
        return kwargs

    def form_valid(self, form):
        action = form.save()
        send_action_mails(action, editor=form.editor)
        messages.success(self.request, self.confirm_message)
        return render(self.request, CLOSE_FRAME, {'url': self.get_success_url()})

    def get_context_data(self, **kwargs):
        """
        :rtype: dict
        """
        ctx = super(ActionView, self).get_context_data(**kwargs)
        form = ctx.get('form')
        # Admin context
        ctx.update({
            'title': self.admin_title,
            'opts': Action._meta,
            'root_path': reverse('admin:index'),
            'adminform': ctx.get('form'),
            'errors': form.errors,
            'is_popup': True,
            'form_url': self.get_form_url(),
            'save_label': self.admin_save_label
        })
        return ctx


class RequestView(ActionView):
    action_type = Action.REQUEST
    admin_title = _('Request approval for changes')
    admin_save_label = _('Request approval for changes')
    confirm_message = _('Successfully requested approval for changes')

    def validate(self):
        super(RequestView, self).validate()
        if self.action_request and not self.action_request.is_closed():
            raise InvalidAction(ACTIVE_REQUEST)


class ApproveRejectMixinView(object):
    def validate(self):
        super(ApproveRejectMixinView, self).validate()
        if self.action_request is None:
            raise InvalidAction(NO_ACTIVE_REQUEST)
        if self.stage is None:
            raise InvalidAction(USER_NOT_ALLOWED)


class ApproveView(ApproveRejectMixinView, ActionView):
    action_type = Action.APPROVE
    admin_title = _('Approve request')
    admin_save_label = admin_title
    confirm_message = _('Request successfully approved')


class RejectView(ApproveRejectMixinView, ActionView):
    action_type = Action.REJECT
    admin_title = _('Reject request')
    admin_save_label = admin_title
    confirm_message = _('Request successfully rejected')


class CancelView(ActionView):
    action_type = Action.CANCEL
    admin_title = _('Cancel request')
    admin_save_label = admin_title
    confirm_message = _('Request successfully cancelled')


class DiffView(TemplateView):
    template_name = 'workflows/admin/action_diff.html'
    pk = None
    language = None

    def render_placeholder(self, placeholder, context):
        return placeholder.render(context, None, editable=False, use_cache=False, lang=self.language)

    def render_page_placeholders(self, page, request):
        new_request = copy(request)
        setattr(new_request, 'current_page', page)

        content_renderer = ContentRenderer(new_request)

        context = SekizaiContext()
        context.update({
            'lang': self.language,
            'current_page': page,
            'request': copy(request),
            'cms_content_renderer': content_renderer,
        })

        return {
            placeholder.slot: self.render_placeholder(placeholder, context) for placeholder in page.placeholders.all()
        }

    def get(self, request, *args, **kwargs):
        self.pk = args[0]
        self.language = args[1]

        return super(DiffView, self).get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(DiffView, self).get_context_data(**kwargs)

        page = get_object_or_404(Page, pk=self.pk)
        public_page = self.render_page_placeholders(page.get_public_object(), self.request)
        draft_page = self.render_page_placeholders(page.get_draft_object(), self.request)

        diffs = []
        for slot, public_rendered in public_page.items():
            draft_rendered = draft_page.pop(slot, [])

            diff = htmldiff(public_rendered, draft_rendered)
            tree = parse_html(diff, cleanup=False)

            for item in tree.xpath("//ins | //del"):
                if len(item):
                    continue

                content = item.text
                if not (content and content.strip()):
                    item.getparent().remove(item)

            diffs.append(etree.tostring(tree, method='html'))

        context.update({
            'title': _('Show current changes'),
            'diffs': diffs,
        })

        return context


WORKFLOW_VIEWS = {
    Action.REQUEST: RequestView,
    Action.APPROVE: ApproveView,
    Action.REJECT: RejectView,
    Action.CANCEL: CancelView,
    Action.DIFF: DiffView,
}


class InvalidAction(Exception):
    def __init__(self, message):
        self.message = message
        super(InvalidAction, self).__init__()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from cms.operations import PUBLISH_PAGE_TRANSLATION
from cms.signals import post_obj_operation
from django.dispatch import receiver

from ..models import Action


# @receiver(post_publish)  # cannot easily get user from this signal unfortunately
@receiver(post_obj_operation)
def close_moderation_request(sender, request=None, operation=None, translation=None, successful=None, **kwargs):
    current_request = Action.get_current_request(translation)

    if all((operation == PUBLISH_PAGE_TRANSLATION, successful, current_request)):
        if not current_request.is_publishable():
            raise ValueError('Page is not publishable!')

        current_action = current_request.last_action()
        current_request.last_action().add_child(
            title=translation,
            workflow=current_action.workflow,
            action_type=Action.PUBLISH,
            user=request.user,
            message=''
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from cms.cms_toolbars import PlaceholderToolbar, PageToolbar, PAGE_MENU_IDENTIFIER
from cms.toolbar.items import ModalButton, Dropdown, DropdownToggleButton, SideframeButton, BaseItem, ButtonList
from cms.toolbar_base import CMSToolbar
from cms.toolbar_pool import toolbar_pool
from cms.extensions.toolbar import ExtensionToolbar
from cms.utils.urlutils import admin_reverse
from cms.utils import get_language_list  # needed to get the page's languages
from django.utils import translation
from django.utils.translation import ugettext_lazy as _

from .models import Action, Workflow, WorkflowExtension


def get_placeholder_toolbar():
    try:
        return toolbar_pool.toolbars.pop('cms.cms_toolbars.PlaceholderToolbar')
    except KeyError:
        return PlaceholderToolbar


def get_page_toolbar():
    try:
        return toolbar_pool.toolbars.pop('cms.cms_toolbars.PageToolbar')
    except KeyError:
        return PageToolbar


class WorkflowExtensionToolbar(ExtensionToolbar):
    # As described in the docs
    model = WorkflowExtension

    def populate(self):
        # setup the extension toolbar with permissions and sanity checks
        current_page_menu = self._setup_extension_toolbar()

        # if it's all ok
        if current_page_menu and self.toolbar.edit_mode:
            # create a sub menu labelled "Workflows" at position 1 in the menu
            sub_menu = self._get_sub_menu(
                current_page_menu, 'submenu_label', _('Workflow'), position=1
            )

            # retrieves the instances of the current title extension (if any) and the toolbar item URL
            urls = self.get_title_extension_admin()

            # we now also need to get the title set (i.e. different language titles) for this page
            page = self._get_page()
            titles = page.title_set.filter(language__in=get_language_list(page.site_id))

            # cycle through the list
            for (title_extension, url), title in zip(urls, titles):

                # adds toolbar items
                sub_menu.add_modal_item(
                    _('for "{title}" ({lang})').format(title=title.title, lang=title.language),
                    url=url,
                    disabled=not self.toolbar.edit_mode
                )


class WorkflowPlaceholderToolbar(get_placeholder_toolbar()):
    def __init__(self, *args, **kwargs):
        super(WorkflowPlaceholderToolbar, self).__init__(*args, **kwargs)
        self.editable = True

    def init_from_request(self):
        super(WorkflowPlaceholderToolbar, self).init_from_request()
        if self.page:
            self.editable = Action.is_editable(self.page.title_set.filter(language=self.current_lang).first())
            self.toolbar.content_renderer._placeholders_are_editable &= self.editable

    def add_structure_mode(self):
        if self.editable:
            return super(WorkflowPlaceholderToolbar, self).add_structure_mode()


# class WorkflowPageToolbar(get_page_toolbar()):
class WorkflowPageToolbar(PageToolbar):
    WORKFLOW_URL_NAME = 'workflow_{}'
    BUTTON_NAMES = {
        Action.REQUEST: _('Request approval for changes'),
        Action.APPROVE: _('Approve changes'),
        Action.REJECT: _('Reject changes'),
        Action.CANCEL: _('Cancel request'),
        Action.DIFF: _('Diff view'),
    }
    current_request = None

    def add_page_menu(self):
        if not self.editable or self.in_app:
            self.toolbar.get_or_create_menu(
                PAGE_MENU_IDENTIFIER, _('Page'), position=1, disabled=True
            )
        else:
            super(WorkflowPageToolbar, self).add_page_menu()

    def init_from_request(self):
        super(WorkflowPageToolbar, self).init_from_request()
        if self.page:
            self.title = self.page.title_set.filter(language=self.current_lang).first()
            self.workflow = Workflow.get_workflow(self.title)
            self.current_request = Action.get_current_request(self.title)
            self.current_action = Action.get_current_action(self.title)
            self.user = self.request.user
            self.next_stage = self.current_action.get_next_stage(self.user) if self.current_action else None
            self.editable = Action.is_editable(self.title)
            self.in_app = self.in_apphook() and not self.in_apphook_root()

    def has_publish_permission(self):
        if getattr(self, 'workflow', None):
            if self.current_request is None or not self.current_request.is_publishable():
                return False
        return super(WorkflowPageToolbar, self).has_publish_permission()

    def has_compare_permission(self):
        if not self.has_dirty_objects():
            return False

        return True

    def has_permission(self, action_type):
        if getattr(self, 'workflow', None) is None:
            return False
        if action_type == Action.CANCEL:
            return self.current_request is not None and not self.current_request.is_closed()
        if action_type == Action.REQUEST:
            if not self.has_dirty_objects():
                return False
            return self.current_request is None or self.current_request.is_closed()
        if action_type in (Action.APPROVE, Action.REJECT):
            return self.current_action and self.current_action.get_next_stage(self.user)
        if action_type == Action.DIFF:
            return self.has_dirty_objects()
        raise ValueError('Unknown action_type: {}'.format(action_type))

    def _button(self, action_type):
        with translation.override(self.current_lang):
            url = admin_reverse(
                self.WORKFLOW_URL_NAME.format(action_type),
                args=[self.page.pk, self.current_lang],
            )
        return ModalButton(name=self.BUTTON_NAMES[action_type], url=url)

    def add_button(self, menu, action_type):
        if self.has_permission(action_type):
            menu.buttons.append(self._button(action_type))

    def add_action_admin_button(self, menu):
        if self.current_request:
            opts = Action._meta
            button = SideframeButton(
                name=_('Show current request in admin'),
                url=admin_reverse(
                    '{app_label}_{model_name}_change'.format(
                        app_label=opts.app_label,
                        model_name=opts.model_name
                    ),
                    args=[self.current_request.pk])
            )
            menu.buttons.append(button)

    def post_template_populate(self):
        self.init_placeholders()
        self.add_draft_live()
        self.add_publish_menu()
        self.add_compare_button()

    def add_publish_button(self, classes=('cms-btn-action', 'cms-btn-publish',)):
        # only do dirty lookups if publish permission is granted else button isn't added anyway
        if self.toolbar.edit_mode and self.has_publish_permission():
            button = self.get_publish_button(classes=classes)
            self.toolbar.add_item(button)

    def add_compare_button(self):
        # only do dirty lookups if compare permission is granted else button isn't added
        if self.toolbar.edit_mode and self.has_compare_permission():
            button_list = ButtonList(side=self.toolbar.RIGHT)
            self.add_button(button_list, Action.DIFF)
            self.toolbar.add_item(button_list)

    def add_publish_menu(self, classes=('cms-btn-action', 'cms-btn-publish', 'cms-btn-publish-active',)):
        if self.in_app:
            return
        workflow_dropdown = Dropdown(side=self.toolbar.RIGHT)
        workflow_dropdown.add_primary_button(
            DropdownToggleButton(name=_('Publish'))
        )
        if self.has_publish_permission():
            workflow_dropdown.buttons.extend(self.get_publish_button(
                classes=('cms-btn-action', 'cms-btn-publish', 'cms-btn-publish-active')).buttons)
        for action_type in (Action.REQUEST, Action.APPROVE, Action.REJECT, Action.CANCEL):
            self.add_button(workflow_dropdown, action_type)
        self.add_action_admin_button(workflow_dropdown)
        if workflow_dropdown.buttons:
            self.toolbar.add_item(workflow_dropdown)


class EditorToolbar(CMSToolbar):
    def populate(self):
        action_dropdown = Dropdown(side=self.toolbar.RIGHT,)
        action_dropdown.add_primary_button(
            DropdownToggleButton(name=_('Pending your approval'))
        )
        actions = Action.requiring_action(self.request.user)
        if actions:
            opts = Action._meta
            view = '{app_label}_{model_name}_change'.format(
                app_label=opts.app_label,
                model_name=opts.model_name
            )
            for a in actions:
                button = SideframeButton(
                    name=str(a.title),
                    url=admin_reverse(view, args=[a.get_request().pk])
                )
                action_dropdown.buttons.append(button)
            self.toolbar.add_item(action_dropdown)


toolbar_pool.register(EditorToolbar)
toolbar_pool.register(WorkflowExtensionToolbar)
toolbar_pool.toolbars['cms.cms_toolbars.PlaceholderToolbar'] = WorkflowPlaceholderToolbar
toolbar_pool.toolbars['cms.cms_toolbars.PageToolbar'] = WorkflowPageToolbar



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from adminsortable2.admin import SortableInlineAdminMixin
from cms.admin.pageadmin import PageAdmin
from cms.extensions.admin import TitleExtensionAdmin
from cms.models import Page, Title
from django.conf.urls import url
from django.contrib import admin, messages
from django.shortcuts import get_object_or_404, redirect
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from .models import WorkflowExtension, Action,WorkflowStage, Workflow
from .views import WORKFLOW_VIEWS


# Register your models here.
# def get_page_admin():
#     pa = admin.site._registry.pop(Page, PageAdmin)
#     print(pa)
#     return pa.__class__


class WorkflowStageInline(SortableInlineAdminMixin, admin.TabularInline):
    model = WorkflowStage

    def get_extra(self, request, obj=None, **kwargs):
        return not (obj and obj.pk)


class WorkflowAdmin(admin.ModelAdmin):
    inlines = [WorkflowStageInline]
    fields = ['name', 'default']
    list_display = ['name', 'default', 'list_stages']

    class Media:
        # fixes display bug in sortable admin tabular inline
        css = {
            'all': ('workflows/css/sortable_inline.css',)
        }

    def list_stages(self, obj):
        return mark_safe(' '.join(self._stage_display(stage) for stage in obj.stages.all()))
    list_stages.short_description = _('Stages')

    def _stage_display(self, stage):
        s = stage.group.name
        if stage.order > 1:
            s = '&rarr; {}'.format(s)
        if stage.optional:
            s = '[{}]'.format(s)
        return s

    def get_queryset(self, request):
        qs = super(WorkflowAdmin, self).get_queryset(request)
        qs = qs.prefetch_related('stages')
        return qs


class WorkflowExtensionAdmin(TitleExtensionAdmin):
    pass


class WorkflowPageAdmin(PageAdmin):
    WORKFLOW_URL_PATTERN = r'^([0-9]+)/([a-z\-]+)/wf/{}/$'
    WORKFLOW_URL_NAME = 'workflow_{}'

    OPEN_REQUEST_MESSAGE = _('Current changes have not yet been approved.')
    NOT_REQUESTED_MESSAGE = _('Current changes must be submitted for approval.')

    def get_urls(self):
        urls = []
        # add workflow action views
        for action_type, view in WORKFLOW_VIEWS.items():
            pattern = self.WORKFLOW_URL_PATTERN.format(action_type)
            name = self.WORKFLOW_URL_NAME.format(action_type)
            urls.append(url(pattern, self.admin_site.admin_view(view.as_view()), name=name))
        return urls + super(WorkflowPageAdmin, self).get_urls()

    def publish_page(self, request, page_id, language):
        title = get_object_or_404(Title, page_id=page_id, language=language, publisher_is_draft=True)
        workflow = Workflow.get_workflow(title)
        current_request = Action.get_current_request(title)

        # legal publishing scenarios: no workflow defined for title OR current request is open and approved
        if not workflow or (current_request and current_request.is_publishable()):
            return super(WorkflowPageAdmin, self).publish_page(request, page_id, language)

        # illegal publishing scenarios: current request not approved or no request at all
        if current_request:
            messages.warning(request, self.OPEN_REQUEST_MESSAGE)
        else:
            messages.warning(request, self.NOT_REQUESTED_MESSAGE)

        return redirect(title.page.get_absolute_url(language, fallback=True))


class ActionAdmin(admin.ModelAdmin):
    # custom templates
    change_form_template = 'workflows/admin/action_change_form.html'
    change_list_template = 'workflows/admin/action_change_list.html'

    readonly_fields = ['title', 'workflow', 'status_display', 'page_link', 'requires_action']
    list_display = ['__str__', 'title', 'status_display', 'created', 'requires_action', 'page_link']
    fieldsets = (
        (None, {
            'fields': (('title', 'workflow', 'status_display', 'requires_action'),),
            'classes': ('wide',)
        }),
        (_('Actions'), {
            'fields': ('page_link',)
        })
    )

    # newest requests first
    ordering = ['-created']

    class Media:
        css = {
            'all': ('workflows/css/action_admin.css',)
        }

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_queryset(self, request):
        self.request = request
        qs = super(ActionAdmin, self).get_queryset(request)
        qs = qs.filter(depth=1)
        return qs

    def requires_action(self, instance):
        return self.request.user in instance.last_action().next_mandatory_stage_editors()
    requires_action.short_description = _('Requires your action')
    requires_action.boolean = True

    def page_link(self, instance):
        t = instance.title
        return mark_safe('<a href="{link}" target="_top" class="close-sideframe-link">{label}</a>'.format(
            link=t.page.get_draft_url(language=t.language, fallback=False),
            label=_('View page')
        ))
    page_link.short_description = _('View page in browser')

    def change_view(self, request, object_id, form_url='', extra_context=None):
        extra = extra_context or {}
        extra.update(self.extra_context(request, object_id))
        return super(ActionAdmin, self).change_view(request, object_id, form_url=form_url, extra_context=extra)

    def extra_context(self, request, object_id):
        action = self.get_object(request, object_id)
        actions = Action.get_tree(parent=action).order_by('depth')
        return {'actions': actions}


if admin.site.is_registered(Page):
    admin.site.unregister(Page)

admin.site.register(Page, WorkflowPageAdmin)
admin.site.register(Workflow, WorkflowAdmin)
admin.site.register(WorkflowExtension, WorkflowExtensionAdmin)
admin.site.register(Action, ActionAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.db.models.deletion
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0006_require_contenttypes_0002'),
        # ('cms', '0018_auto_20170825_1058'),
        ('cms', '0016_auto_20160608_1535'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Action',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('path', models.CharField(unique=True, max_length=255)),
                ('depth', models.PositiveIntegerField()),
                ('numchild', models.PositiveIntegerField(default=0)),
                ('action_type', models.CharField(choices=[('request', 'request'), ('approve', 'approve'), ('reject', 'reject'), ('cancel', 'cancel'), ('publish', 'publish'), ('diff', 'diff')], default='request', max_length=10, verbose_name='Action type')),
                ('created', models.DateTimeField(verbose_name='Created', auto_now_add=True)),
                ('message', models.TextField(verbose_name='Message')),
                ('group', models.ForeignKey(default=None, on_delete=django.db.models.deletion.PROTECT, verbose_name='Group', null=True, to='auth.Group')),
            ],
            options={
                'verbose_name': 'Workflow action',
                'verbose_name_plural': 'Workflow actions',
                'ordering': ('depth', 'created'),
            },
        ),
        migrations.CreateModel(
            name='Workflow',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('name', models.CharField(verbose_name='Name', unique=True, help_text='Please provide a descriptive name!', max_length=63)),
                ('default', models.BooleanField(default=False, help_text='Should this be the default workflow for all pages and languages?', verbose_name='Default workflow')),
            ],
            options={
                'verbose_name': 'Workflow',
                'verbose_name_plural': 'Workflows',
            },
        ),
        migrations.CreateModel(
            name='WorkflowExtension',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('descendants', models.BooleanField(default=True, help_text='Should this workflow apply to descendant pages?', verbose_name='Descendants')),
                ('extended_object', models.OneToOneField(editable=False, to='cms.Title')),
                ('public_extension', models.OneToOneField(related_name='draft_extension', null=True, editable=False, to='workflows.WorkflowExtension')),
                ('workflow', models.ForeignKey(help_text='The workflow set here is language specific.', verbose_name='Workflow', to='workflows.Workflow')),
            ],
            options={
                'verbose_name': 'Workflow extension',
                'verbose_name_plural': 'Workflow extensions',
            },
        ),
        migrations.CreateModel(
            name='WorkflowStage',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('order', models.PositiveSmallIntegerField(verbose_name='Order', default=0)),
                ('optional', models.BooleanField(default=False, help_text='Is this workflow stage optional?', verbose_name='Optional')),
                ('group', models.ForeignKey(help_text='Only members of this group can approve this workflow stage.', on_delete=django.db.models.deletion.PROTECT, verbose_name='Group', to='auth.Group')),
                ('workflow', models.ForeignKey(related_name='stages', verbose_name='Workflow', to='workflows.Workflow')),
            ],
            options={
                'verbose_name': 'Workflow stage',
                'verbose_name_plural': 'Workflow stages',
                'ordering': ('order',),
            },
        ),
        migrations.AddField(
            model_name='action',
            name='stage',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.SET_NULL, verbose_name='Stage', null=True, to='workflows.WorkflowStage'),
        ),
        migrations.AddField(
            model_name='action',
            name='title',
            field=models.ForeignKey(verbose_name='Title', to='cms.Title'),
        ),
        migrations.AddField(
            model_name='action',
            name='user',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, verbose_name='User', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='action',
            name='workflow',
            field=models.ForeignKey(verbose_name='Workflow', to='workflows.Workflow'),
        ),
        migrations.AlterUniqueTogether(
            name='workflowstage',
            unique_together=set([('workflow', 'group')]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django import forms
from django.contrib.auth import get_user_model
from django.utils.translation import ugettext_lazy as _

from .models import Action


class ActionForm(forms.Form):
    message_ = forms.CharField(
        label=_('Message'),
        required=False,
        help_text=_('You may provide some more information.'),
        widget=forms.Textarea
    )

    editor_ = forms.ModelChoiceField(
        label=_('Editor'),
        queryset=get_user_model().objects.none(),
        help_text=_('Only notify a specific user?'),
        required=False
    )

    def __init__(self, *args, **kwargs):
        self.stage = kwargs.pop('stage', None)
        self.title = kwargs.pop('title')
        self.request = kwargs.pop('request')
        self.workflow = kwargs.pop('workflow')
        self.action_type = kwargs.pop('action_type')  # {open, approve, reject, cancel}
        self.next_stage = self.workflow.next_mandatory_stage(self.stage)
        self.group = getattr(self.stage, 'group', None)
        cr = Action.get_current_request(self.title)
        self.current_action = None if (not cr or cr.is_closed()) else cr.last_action()
        self.user = self.request.user
        super(ActionForm, self).__init__(*args, **kwargs)
        self.adjust_editor()

    @property
    def message(self):
        """
        :rtype: str
        """
        return self.cleaned_data.get('message_', '')

    @property
    def editor(self):
        """
        :rtype: django.contrib.auth.models.User
        """
        return self.cleaned_data.get('editor')

    @property
    def editors(self):
        if self.next_stage is None:
            raise ValueError('No next stage!')
        if self.editor:
            return get_user_model().objects.filter(pk=self.editor.pk)
        return self.next_stage.group.user_set.all()

    def adjust_editor(self):
        if self.action_type in (Action.CANCEL, Action.REJECT) or self.next_stage is None:
            self.fields.pop('editor_', None)  # no editor can be chosen
            return
        group = self.next_stage.group
        self.fields['editor_'].queryset = group.user_set.all()
        self.fields['editor_'].empty_label = _('Any {}').format(group.name)

    def save(self):
        """
        :rtype: Action
        """
        init_kwargs = {
            attr: getattr(self, attr) for attr in
            ('message', 'user', 'title', 'workflow', 'stage', 'action_type', 'group')
        }
        if self.current_action is None:
            assert self.action_type == Action.REQUEST  # root must be request
            return Action.add_root(**init_kwargs)
        else:
            assert self.action_type != Action.REQUEST  # non-root must not be request
            return self.current_action.add_child(**init_kwargs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class WorkflowsConfig(AppConfig):
    name = 'workflows'
    verbose_name = _('Django CMS Workflows')

    def ready(self):
        # needed to connect signal handlers
        from workflows.signals import handlers



#!/usr/bin/env python
import os
import uuid
from setuptools import find_packages, setup

from pip.req import parse_requirements

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

requirements = [str(ir.req) for ir in parse_requirements('requirements.txt', session=uuid.uuid1())]

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

setup(
    name='djangocms-workflows',
    version='0.2a1',
    description='Workflow system for Django-CMS.',
    long_description=README,
    license=read('LICENSE'),
    author='karalics',
    author_email='slavisa.karalic@gmail.com',
    url='https://github.com/karalics/djangocms-workflows.git',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    keywords=['django', 'Django CMS', 'workflow', 'bootstrap', 'website', 'CMS'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging

from cms.extensions.extension_pool import extension_pool
from cms.extensions.models import TitleExtension
from cms.models import Title
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from treebeard.mp_tree import MP_Node

logger = logging.getLogger('django.cms-workflows')


class Workflow(models.Model):
    """
    Model representing a editorial workflow. A workflow is simply a sequence of ordered
    mandatory or optional stages. Every `cms.Title` instance can be associated with a
    concrete Workflow instance via the WorkflowExtension. At most one workflow can be
    made the default workflow which is assumed for all titles without their own or
    inherited workflow.
    """
    name = models.CharField(
        _('Name'),
        max_length=63,
        unique=True,
        help_text=_('Please provide a descriptive name!')
    )

    default = models.BooleanField(
        _('Default workflow'),
        default=False,
        help_text=_('Should this be the default workflow for all pages and languages?')
    )

    class Meta:
        verbose_name = _('Workflow')
        verbose_name_plural = _('Workflows')

    def save(self, **kwargs):
        if self.default:
            Workflow.objects.filter(default=True).exclude(pk=self.pk).update(default=False)
        super(Workflow, self).save(**kwargs)

    def __str__(self):
        return self.name

    @classmethod
    def default_workflow(cls):
        """
        :return: The default workflow if one exists, else `None`
        """
        try:
            return cls.objects.get(default=True)
        except cls.DoesNotExist:
            return None
        except cls.MultipleObjectsReturned:
            logger.error('Multiple default workflows set. This should not happen!')
            raise

    @classmethod
    def get_workflow(cls, title):
        """Returns appropriate workflow for this title.

        :type title: Title
        :rtype: Workflow | None
        :raises: Workflow.MultipleObjectsReturned
        """
        if title is None:
            return cls.default_workflow()
        # 1. check for custom workflow
        try:
            return title.workflowextension.workflow
        except WorkflowExtension.DoesNotExist:
            pass

        # 2. check for inherited workflow up the site tree
        titles = Title.objects.filter(page__in=title.page.get_ancestors())  # all title of ancestor pages ...
        titles = titles.filter(language=title.language)  # and same language ...
        titles = titles.filter(workflowextension__descendants=True)  # that have an inherited workflow ...
        titles = titles.order_by('-page__depth')  # bottom up
        ancestor_title = titles.first()

        if ancestor_title:
            return ancestor_title.workflowextension.workflow

        # 3. check for default workflow, might be None
        return cls.default_workflow()

    @cached_property
    def mandatory_stages(self):
        return self.stages.filter(optional=False)

    @cached_property
    def first_mandatory_stage(self):
        return self.mandatory_stages.first()

    def possible_next_stages(self, stage=None):
        """
        Return a queryset of all possible stages that can be used for the next action. These
        are all stages starting with the very next stage up to and including the next mandatory
        stage.
        """
        if stage:
            return stage.possible_next_stages
        fms = self.first_mandatory_stage
        return self.stages.filter(order__lte=fms.order)

    def next_mandatory_stage(self, stage=None):
        if stage is None:
            return self.first_mandatory_stage
        return self.mandatory_stages.filter(order__gt=stage.order).first()


class WorkflowStage(models.Model):
    """
    Represents a stage of a workflow. It is associated with a group. That means only a member of the
    group can approve or reject the stage.
    """
    workflow = models.ForeignKey(
        'workflows.Workflow',
        on_delete=models.CASCADE,
        verbose_name=_('Workflow'),
        related_name='stages'
    )

    group = models.ForeignKey(
        'auth.Group',
        on_delete=models.PROTECT,
        verbose_name=_('Group'),
        help_text=_('Only members of this group can approve this workflow stage.'),
    )

    # admin-sortable required
    order = models.PositiveSmallIntegerField(
        _('Order'),
        default=0
    )

    optional = models.BooleanField(
        _('Optional'),
        default=False,
        help_text=_('Is this workflow stage optional?')
    )

    class Meta:
        verbose_name = _('Workflow stage')
        verbose_name_plural = _('Workflow stages')
        ordering = ('order',)
        unique_together = (('workflow', 'group'),)

    def __str__(self):
        return self.group.name + ('[optional]' if self.optional else '')

    @cached_property
    def next_mandatory_stage(self):
        return self.workflow.stages.filter(order__gt=self.order, optional=False).first()

    @cached_property
    def possible_next_stages(self):
        nms = self.next_mandatory_stage
        pending = self.workflow.stages.filter(order__gt=self.order)
        if nms is not None:
            pending = pending.filter(order__lte=nms.order)
        return pending


class WorkflowExtension(TitleExtension):
    """
    This extension holds the workflow association of a cms.Title instance.
    """
    workflow = models.ForeignKey(
        'workflows.Workflow',
        on_delete=models.CASCADE,
        verbose_name=_('Workflow'),
        help_text=_('The workflow set here is language specific.')
    )

    descendants = models.BooleanField(
        _('Descendants'),
        help_text=_('Should this workflow apply to descendant pages?'),
        default=True
    )

    class Meta:
        verbose_name = _('Workflow extension')
        verbose_name_plural = _('Workflow extensions')

    def __str__(self):
        return '{title} ({lang}): {workflow}'.format(
            title=self.extended_object.title,
            lang=self.extended_object.language,
            workflow=self.workflow
        )

    @cached_property
    def language(self):
        return self.extended_object.language


extension_pool.register(WorkflowExtension)


class Action(MP_Node):
    """
    Actions are the instantiations of workflow stages. They model a concrete editorial process that
    is modelled as a linked sequence (a 'unary' tree) using treebeard's MP_Node:

    The initial submission request is always a root node of action_type REQUEST.
    This is followed by n actions of type APPROVE each of which is associated with a stage of the
    appropriate workflow.
    The last action in such a chain is an action of either of REJECT, CANCEL, or PUBLISH.
    CANCEL/PUBLISH actions are not associated with a stage, but REJECT is.

    E.g.:

    workflow: editor1 [-> editor2] -> test_user    (editor2 is optional)

        action: stage/group
    ==========:============
    1. REQUEST: --
    2. APPROVE: editor1
    3. APPROVE: editor2
    4. APPROVE: test_user
    5. PUBLISH: --

    1. REQUEST: --
    2. APPROVE: editor1
    3. APPROVE: test_user
    4. PUBLISH: --

    1. REQUEST: --
    2. APPROVE: editor1
    3.  REJECT: test_user

    1. REQUEST: --
    2. APPROVE: editor1
    3.  CANCEL: --
    """
    REQUEST, APPROVE, REJECT, CANCEL, PUBLISH, DIFF = 'request', 'approve', 'reject', 'cancel', 'publish', 'diff'
    TYPES = (
        (REQUEST, _('request')),
        (APPROVE, _('approve')),
        (REJECT, _('reject')),
        (CANCEL, _('cancel')),
        (PUBLISH, _('publish')),
        (DIFF, _('diff')),
    )

    REQUESTED, APPROVED, REJECTED, CANCELLED, PUBLISHED = 'requested', 'approved', 'rejected', 'cancelled', 'published'
    STATUS = (
        (REQUESTED, _('Requested')),
        (APPROVED, _('Approved')),
        (REJECTED, _('Rejected')),
        (CANCELLED, _('Cancelled')),
        (PUBLISHED, _('Published')),
    )

    title = models.ForeignKey(
        'cms.Title',
        on_delete=models.CASCADE,
        verbose_name=_('Title'),
    )

    # persist workflow in case the workflow changes on the title
    workflow = models.ForeignKey(
        'workflows.Workflow',
        on_delete=models.CASCADE,
        verbose_name=_('Workflow'),
    )

    stage = models.ForeignKey(
        'workflows.WorkflowStage',
        on_delete=models.SET_NULL,
        verbose_name=_('Stage'),
        # allow null for open action and if Stage is deleted on Workflow
        null=True,
        default=None,
    )

    # persist stage Group in case stage is deleted
    group = models.ForeignKey(
        'auth.Group',
        on_delete=models.PROTECT,
        verbose_name=_('Group'),
        null=True,
        default=None
    )

    action_type = models.CharField(
        _('Action type'),
        max_length=10,
        choices=TYPES,
        default=REQUEST,
    )

    created = models.DateTimeField(
        _('Created'),
        auto_now_add=True,
    )

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        verbose_name=_('User'),
        null=True
    )

    message = models.TextField(
        _('Message'),
    )

    class Meta:
        verbose_name = _('Workflow action')
        verbose_name_plural = _('Workflow actions')
        ordering = ('depth', 'created')

    def __str__(self):
        return '#{}: {}'.format(self.title_id, self.action_type)

    def save(self, **kwargs):
        if self.action_type == self.REQUEST:
            try:
                previous = Action.get_requests(title=self.title).exclude(pk=self.pk).latest('created')
            except Action.DoesNotExist:
                pass
            else:
                assert previous.is_closed(), 'Close previous request #{pk} before new request'.format(pk=previous.pk)
        super(Action, self).save(**kwargs)

    def is_closed(self):
        return self.last_action().action_type in (self.REJECT, self.CANCEL, self.PUBLISH)

    def get_request(self):
        """Root of this chain of actions.

        :rtype: Action
        :return:
        """
        return self.get_root()

    def get_author(self):
        """Return author of changes.

        :rtype: django.contrib.auth.models.AbstractUser
        :return: author of changes
        """
        return self.get_request().user

    def next_mandatory_stage(self):
        """
        :rtype: WorkflowStage
        :return:
        """
        if self.action_type == self.REQUEST:
            return self.workflow.first_mandatory_stage
        if self.action_type == self.APPROVE and self.stage:
            return self.stage.next_mandatory_stage
        return None

    def next_mandatory_stage_editors(self):
        nms = self.next_mandatory_stage()
        if not nms:
            return get_user_model().objects.none()
        return nms.group.user_set.all()

    def possible_next_stages(self):
        return self.workflow.possible_next_stages(self.stage)

    def last_action(self):
        """
        Returns the latest action of this action's action chain.

        :rtype: Action
        """
        return Action.get_tree(parent=self).latest('depth')

    def is_publishable(self):
        """
        Is this action's title publishable with regard to this action's action chain?

        :rtype: bool
        """
        if self.is_closed():  # rejected, cancelled, or already published?
            return False
        last_action = self.last_action()
        if last_action.action_type != self.APPROVE:
            return False
        return last_action.next_mandatory_stage() is None

    def get_next_stage(self, user):
        if self.is_closed():
            return None
        return self.possible_next_stages().filter(group__user=user).last()

    @cached_property
    def status(self):
        la = self.last_action()
        if la.action_type == self.CANCEL:
            return self.CANCELLED
        if la.action_type == self.PUBLISH:
            return self.PUBLISHED
        if la.action_type == self.REJECT:
            return self.REJECTED
        if self.is_publishable():
            return self.APPROVED
        else:
            return self.REQUESTED

    @cached_property
    def status_display(self):
        return dict(self.STATUS)[self.status]
    status_display.short_description = _('Status')

    @classmethod
    def get_requests(cls, title=None):
        requests = cls.get_root_nodes()
        if title:
            requests = requests.filter(title=title)
        return requests

    @classmethod
    def get_current_request(cls, title):
        """
        Returns the most recent request (root action) for this title.

        :rtype: Action
        """
        workflow = Workflow.get_workflow(title)
        if workflow is None:
            return None
        try:
            # there can only be one open request per title at a time and it must be the last
            return cls.get_requests(title=title).latest('created')
        except cls.DoesNotExist:
            return None

    @classmethod
    def get_current_action(cls, title):
        """
        Returns the most recent action for this title.

        :type title: Title
        :param title:
        :rtype: Action
        """
        latest_request = cls.get_current_request(title)
        if latest_request:
            return cls.get_tree(parent=latest_request).latest('depth')
        return None

    @classmethod
    def is_editable(cls, title):
        """
        Can this title be edited at the moment?
        Only returns `True` if there is no open request at the moment.
        :type title: Title
        :rtype: bool
        """
        if not title.publisher_is_draft:
            return False
        current_request = cls.get_current_request(title)
        if current_request is None:
            return True
        return current_request.is_closed()

    @classmethod
    def requiring_action(cls, user):
        """
        Returns a list of all actions that currently require approve or reject activity
        by this user.

        :type user: django.contrib.auth.models.User
        :rtype: list
        """
        actions = []
        for title in Title.objects.filter(action__isnull=False).distinct():
            ca = cls.get_current_action(title)
            if user in ca.next_mandatory_stage_editors():
                actions.append(ca)
        return actions
